/*
 *	gxcardmon.c
 *	Copyright (C) 1999 by Yoshishige Arai <ryo2@jp.freebsd.org>
 * All right reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = "$Id: gxcardmon.c,v 1.4 1999/06/12 05:25:14 ryo2 Exp ryo2 $";
#endif

#ifdef USE_PIXMAP
#include "power.xpm"
#define PIXMAP_NAME power
#endif

#define VERSION "0.5"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <pccard/cardinfo.h>
#include <pccard/cis.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <regex.h>

#include <gtk/gtk.h>
#include <X11/Xlib.h>

#define MY_FRAME_WIDTH		250

#ifndef CONFIG_FILE
#define CONFIG_FILE		"/usr/local/lib/xpccard/config"
#endif
#ifndef RC_FILE
#define RC_FILE		"/usr/local/etc/xpccard/gxcardmon.gtkrc"
#endif
#define LABEL_CARDNAME		"cardname"
#define SERVER_NAME		"/var/tmp/.pccardd"
#define CLIENT_NAME		"/tmp/.xpccard"

#define CMD_NR_SLOTS		"S\n"
#define CMD_GET_INFO		"N%d\n"
#define CMD_PWR_ON		"P%d\n"
#define CMD_PWR_OFF		"Q%d\n"

#define MAX_PC_SLOT		4
#define MAX_INFO_LEN		64
#define MAX_CARD_NAME		MAX_INFO_LEN * 2
#define MAX_NAME_FEILD		5
#define S_NA			"N/A"
#define S_EMPTY			"Empty"
#define S_HIBERNATING		"Hibernating Card"

GtkTooltips *tips;
GtkStyle *style;
GdkPixmap *pixm;
GtkWidget *window, *text, *table, *label, *menu, *menubutton, *hbox, *button;
GtkWidget *menuitem, *vbox, *vbox3, *w, *sep;
#ifdef USE_PIXMAP
GtkWidget *pixmapw;
GdkPixmap *pixmap;
GdkBitmap *mask;
#endif

static char *myname;
static int SOCKET;
static int nr_slot = 0;
static int slen, clen;
static struct sockaddr_un saddr, caddr;
static int x_size = 360, y_size = 28;
static Bool sensitive = TRUE, small = FALSE;
static char *config_file = CONFIG_FILE;
static char *rc_file = RC_FILE;

struct CardInfo {
    int  stat;
    char manuf[MAX_INFO_LEN+4];
    char vers[MAX_INFO_LEN+4];
    char drv[MAX_INFO_LEN+4];
    char name[MAX_CARD_NAME+4];
};
static struct CardInfo cardInfo[MAX_PC_SLOT];

static struct io_desc cardIO[MAX_PC_SLOT];

struct CardAlias {
    char *alias;
    char *manuf;
    char *vers;
    char *buff;
    struct CardAlias *next;
};
static struct CardAlias *aliasTop = NULL;
static struct CardAlias *aliasPrev;

struct SlotFrame {
    char num[3];
    GtkWidget *hbox, *pwrbutton, *drvbutton, *drvbox, *drvlabel;
    GtkWidget *drvinfo, *label, *sep, *style;
    GtkTooltips *tooltips, *tooltips2;
#ifdef USE_PIXMAP
    GtkWidget *pix;
#endif    
};
static struct SlotFrame slotFrame[MAX_PC_SLOT];

/* prototypes */
static void usage(void);
static int  openSocket(void);
static void closeSocket(void);
static int  communicate(char *, char *, int);
static int  ismatch(char *, char*);
static char *nameAlias(struct CardAlias *, struct CardInfo *);
static int  parseCardName(char *, struct CardInfo *);
static int  getCardIOinfo(int, struct io_desc *);
static int  shellWord(char **, char **);
static int  readConfig(char *);
static void freeAliasArray(struct CardAlias **);
static void getCardInfo(int, Bool);
static void sigHupHandler(int);
static void sigTermHandler(int);
static void statChanged(gpointer, gint, GdkInputCondition);
static void togglePowerState(int);
static void toggle_buttonCB (GtkWidget *,gpointer);
static void accelerate_keyCB (GtkWidget *widget, GdkEventKey *event);
static gint delete_event(GtkWidget *, GdkEvent *, gpointer);
static void destroy (GtkWidget *, gpointer);


static gint delete_event(GtkWidget *widget, GdkEvent *e, gpointer data)
{
    closeSocket();
    freeAliasArray(&aliasTop);
    return (FALSE);
}

static void destroy (GtkWidget *widget, gpointer data)
{
    delete_event(NULL, NULL, NULL);
    gtk_main_quit ();
}

static void usage()
{
    static char *option_help[] = {
	"-s		make buttons always be sensitive",
	"-n		smaller size without card names",
	"-f CONFIG	specify config file",
	"-r GTKRC	specify rc file",
	"-v		show version",
	"-h		this help",
	NULL};
    char **cpp;

    fprintf(stderr, "\nusage: %s [-toolkitoption ...] [options]\noptions:\n",
	    myname);
    for (cpp = option_help; *cpp; cpp++)
        fprintf(stderr, "    %s\n", *cpp);
    exit(0);
}

static int openSocket()
{
    umask(077);
    if ((SOCKET = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
	return(-1);
    bzero((char *)&caddr, sizeof(caddr));
    caddr.sun_family = AF_UNIX;
    sprintf(caddr.sun_path, "%s%06d", CLIENT_NAME, getpid());
    clen = SUN_LEN(&caddr);
    if (bind(SOCKET, (struct sockaddr *)&caddr, clen) < 0)
	return(-1);
    bzero((char *)&saddr, sizeof(saddr));
    saddr.sun_family = AF_UNIX;
    strcpy(saddr.sun_path, SERVER_NAME);
    slen = SUN_LEN(&saddr);
    return(1);
}

static void closeSocket()
{
    close(SOCKET);
    freeAliasArray(&aliasTop);
    unlink(caddr.sun_path);
}

static int communicate(cmd, ansbuf, buflen)
char *cmd, *ansbuf;
int buflen;
{
    struct timeval timeout;
    fd_set fds;
    int n, k;

    n = strlen(cmd);
    if (n) {
	if (sendto(SOCKET, cmd, n, 0, (struct sockaddr *)&saddr, slen) != n) {
	    perror("send");
	    exit(1);
	}
    }

    if (buflen == 0) return(0);

    FD_ZERO(&fds);
    FD_SET(SOCKET, &fds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    n = 0;
    if (select(FD_SETSIZE, &fds, NULL, NULL, &timeout) > 0) {
	if (FD_ISSET(SOCKET, &fds)) {
	    k = 0;
	    if ((n = recvfrom(SOCKET, ansbuf, buflen, 0, NULL, &k)) < 0) {
		perror("recvfrom");
		exit(1);
	    }
	}
    }
    *(ansbuf + n) = 0;
    return(n);
}

static int ismatch(key, str)
char *key, *str;
{
    int r;
    regex_t exp;

    if ((r = regcomp(&exp, key, REG_EXTENDED | REG_NOSUB)) == 0) {
	r = regexec(&exp, str, (size_t)0, NULL, 0);
	regfree(&exp);
    }
    return(r == 0);
}

static char *nameAlias(array, card)
struct CardAlias *array;
struct CardInfo *card;
{
    while(array != NULL) {
	if (ismatch(array->manuf, card->manuf) && ismatch(array->vers, card->vers)) {
	    return(array->alias);
	}
	array = array->next;
    }
    return(NULL);
}

static int parseCardName(buf, card)
char *buf;
struct CardInfo *card;
{
    char *a[MAX_NAME_FEILD];
    char *p, *r, *s;
    int i, slot;

    if ((p = strdup(buf)) == NULL) {
	return(-1);
    }
    s = p + strlen(buf) - 1;
    while(s > p && *s == ' ') *s-- = 0;
    for(s = p, i = 0; i < MAX_NAME_FEILD; i++) {
	while(*s && *s == ' ') s++;
	a[i] = s;
	if (*s == 0) continue;
	while(*s && *s != '~') s++;
	if (i < MAX_NAME_FEILD - 1 && *s == '~') {
	    r = s - 1;
	    *s++ = 0;
	    while(r > a[i] && *r == ' ') *r-- = 0;
	}
    }
    slot = atoi(a[0]);
    card->stat = atoi(a[4]);
    strncpy(card->manuf, *a[1] ? a[1] : S_NA, MAX_INFO_LEN);
    strncpy(card->vers, *a[2] ? a[2] : S_NA, MAX_INFO_LEN);
    strncpy(card->drv, *a[3] ? a[3] : (card->stat == 2 ? "N/A" : S_NA), MAX_INFO_LEN);

    switch (card->stat) {
    case 0:
	strcpy(card->name, S_EMPTY);
	break;
    case 2:
	strcpy(card->name, S_HIBERNATING);
	break;
    default:
	if ((s = nameAlias(aliasTop, card)) == NULL) {
	    snprintf(card->name, MAX_CARD_NAME, "%s (%s)", card->manuf, card->vers);
	} else {
	    strncpy(card->name, s, MAX_CARD_NAME);
	}
	break;
    }
    free(p);
    return(slot);
}

static int getCardIOinfo(slot, card)
int slot;
struct io_desc *card;
{
    int fd;
    char buf[64];

    card->window = slot;
    card->flags = card->start = card->size = 0;
    sprintf(buf, CARD_DEVICE, slot);
    if (0 > (fd = open(buf, O_RDONLY))) return -1;
    if (ioctl(fd, PIOCGIO, card))
	g_message("%s: ioctl (PIOCGIO) error.\n", myname);
    close(fd);
    return 1;
}

static int shellWord(dst, src)
char **dst, **src;
{
    int quote, n;
    char *d, *s;

    d = *dst; s = *src;
    while (*s && (*s == ' ' || *s == '\t')) s++;
    quote = n = 0;
    while (*s) {
	if (quote == 0 && (*s == ' ' || *s == '\t')) break;
	switch(*s) {
	case '"':
	    if (quote)  quote = 0;
	    else	quote = 2;
	    s++;
	    break;
	case '\\':
	    s++;
	    if (*s == 0) break;
	default:
	    *d++ = *s++;  n++;
	    break;
	}
    }
    *d = 0;   *dst = ++d; *src = s;
    return(n);
}

static int readConfig(file)
char *file;
{
    FILE *fp;
    int n, cnt, line, len;
    char buff[1024];
    char *s, *p, *a[10], *dst;
    struct CardAlias *alp;

    if ((fp = fopen(file, "r")) == NULL) {
	return(-1);
    }

    line = cnt = 0;
    while(fgets(buff, sizeof(buff), fp) != NULL) {
	line++;
	*(buff + strlen(buff) - 1) = 0;
	s = buff;
	while(*s && (*s == ' ' || *s == '\t')) s++;
	if (*s == 0 || *s == '#') continue;

	dst = (char *)g_malloc(strlen(s));
	n = 0;
	p = dst;
	len = 0;
	while (*s) {
	    a[n++] = p;
	    len += shellWord(&p, &s);
	}

	if (n == 4 && strcmp(a[0], "alias") == 0) {
	    alp = (struct CardAlias *)g_malloc(sizeof(struct CardAlias));
	    alp->buff  = dst;
	    alp->alias = a[1];
	    alp->manuf = a[2];
	    alp->vers  = a[3];
	    if (aliasTop == NULL)
		aliasTop = alp;
	    else {
		aliasPrev->next = alp;
		alp->next = NULL;
	    }
	    aliasPrev = alp;
	    cnt++;
	} else {
	    g_message("readConfig() #%d: illegal format.\n", line);
	    free(dst);
	    continue;
	}
    }
    fclose(fp);
    return(cnt);
}

static void freeAliasArray(array)
struct CardAlias **array;
{
    struct CardAlias *p, *chan;

    p = *array;
    while(p != NULL) {	
	chan = p->next;
	free(p->buff);
	free(p);
	p = chan;
    }
    *array = NULL;
}

static void getCardInfo(nr_slot, noset_flag)
int nr_slot;
Bool noset_flag;
{
    int i, r;
    char cmd[16], ansbuf[BUFSIZ];

    readConfig(config_file);
    for(i = 0; i < nr_slot; i++) {
	sprintf(cmd, CMD_GET_INFO, i);
	if ((r = communicate(cmd, ansbuf, BUFSIZ)) > 0) {
	    parseCardName(ansbuf, &cardInfo[i]);
	    getCardIOinfo(i, &cardIO[i]);
	}
	if (!noset_flag)
	    gtk_label_set (GTK_LABEL(slotFrame[i].label),
			   ((cardInfo[i].stat==1)? cardInfo[i].name:S_HIBERNATING));
    }
}

static void sigHupHandler(sig)
int sig;
{
    freeAliasArray(&aliasTop);
    getCardInfo(nr_slot, small);
}

static void sigTermHandler(sig)
int sig;
{
    destroy(NULL, NULL);
}

static void statChanged(data, socket, condition)
gpointer data;
gint socket;
GdkInputCondition condition;
{
    int k, n, slot;
    char ansbuf[BUFSIZ];
    struct CardInfo card;

    if (socket <= 0) socket = SOCKET;
    k = 0;
    n = recvfrom(SOCKET, ansbuf, BUFSIZ, 0, NULL, &k);
    if (n == 0) return;

    *(ansbuf + n) = 0;
    slot = parseCardName(ansbuf, &card);
    getCardIOinfo(slot, &cardIO[slot]);
    if (card.stat == 2)
	cardInfo[slot].stat = card.stat;
    else
	cardInfo[slot] = card;

    if (!small)
	gtk_label_set (GTK_LABEL(slotFrame[slot].label),
		       ((cardInfo[slot].stat==1)? cardInfo[slot].name:
			(cardInfo[slot].stat==2)? S_HIBERNATING:S_EMPTY));
    gtk_label_set(GTK_LABEL(slotFrame[slot].drvlabel),
		  ((cardInfo[slot].stat==1)? cardInfo[slot].drv:S_NA));
    if (sensitive && cardInfo[slot].stat==1)
	gtk_widget_set_sensitive(slotFrame[slot].drvbutton, TRUE);
    if (cardInfo[slot].stat != 2)
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (slotFrame[slot].pwrbutton), FALSE);
}

static void togglePowerState(int i)
{
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (slotFrame[i].pwrbutton),
				(GTK_TOGGLE_BUTTON (slotFrame[i].pwrbutton)->active)?
				FALSE:TRUE);
    if (sensitive && ! GTK_TOGGLE_BUTTON (slotFrame[i].pwrbutton)->active)
 	gtk_widget_set_sensitive(slotFrame[i].drvbutton, FALSE);
}

static void toggle_buttonCB (GtkWidget *widget, gpointer data)
{
    int slot;
    char cmd[8];

    slot = atoi((char *)data);

    /* 0:empty  1:active  2:hibernate */
    if (GTK_TOGGLE_BUTTON (widget)->active) {
	/** Toggle ON **/
	if (cardInfo[slot].stat != 1) return;
	sprintf(cmd, CMD_PWR_OFF, slot);
	if (sensitive)
	    gtk_widget_set_sensitive(slotFrame[slot].drvbutton, FALSE);
    } else {
	/** Toggle OFF **/
	if (cardInfo[slot].stat != 2) return;
	/* hibernate */
	if (!small)
	    gtk_label_set (GTK_LABEL(slotFrame[slot].label), "Detecting card info...");
	sprintf(cmd, CMD_PWR_ON, slot);
    }
    communicate(cmd, NULL, 0);
}

static void accelerate_keyCB (GtkWidget *widget, GdkEventKey *event)
{
    switch (event->keyval) {
    case 'q':
	destroy(NULL, NULL);
	return;
    case '0':
    case '1':
    case '2':
    case '3':
 	togglePowerState((event->keyval) -'0');
    default:
	return;
    }
}

int main (int argc, char **argv)
{
    GtkWidget *window;
    gint inputtag;
    int i, r;
    char line[128];
    char ansbuf[BUFSIZ];

    if ((myname = strrchr(*argv, '/')) == NULL)
	myname = *argv;
    else
	myname++;
    for(i=1; argv[i] != NULL; i++) {
	if(!strcmp(argv[i],"-s")) sensitive = FALSE;
	if(!strcmp(argv[i],"-n")) {small = TRUE; x_size = 80; }
	else if(!strcmp(argv[i],"-f")) { config_file = argv[++i]; }
	else if(!strcmp(argv[i],"-r")) { rc_file = argv[++i]; }
	else if(!strcmp(argv[i],"-v")) {
	    fprintf(stderr, "%s version %s (c) 1999 by Yoshishige Arai <ryo2@jp.freebsd.org>\nRCS %s\n", myname, VERSION, rcsid+1);
	    exit(0);
	}
	else if(!strncmp(argv[i],"-h", 2) || !strncmp(argv[i],"--h", 3)) usage();
    }

    gtk_init (&argc, &argv);    
#ifdef I18N
    gtk_set_locale();
#endif
    if (openSocket() < 0) {
	perror("openSocket");
	exit(1);
    }
    if ((r = communicate(CMD_NR_SLOTS, ansbuf, BUFSIZ)) > 0) {
	nr_slot = atoi(ansbuf);
    }
    if (nr_slot == 0) {
	closeSocket();
	exit(0);
    }
    getCardInfo(nr_slot, True);
    gtk_rc_parse(rc_file);
    signal(SIGHUP, sigHupHandler);
    signal(SIGINT, sigTermHandler);
    signal(SIGTERM, sigTermHandler);
    signal(SIGQUIT, sigTermHandler);

    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_usize (GTK_WIDGET (window), x_size, y_size*nr_slot);
#ifdef USE_PIXMAP
    gtk_widget_show (window);
#endif
    gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect (GTK_OBJECT (window), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);
    gtk_signal_connect (GTK_OBJECT (window), "key_press_event",
			GTK_SIGNAL_FUNC (accelerate_keyCB), NULL);

    sprintf(line, "%s-%s\n", myname, VERSION);
    gtk_window_set_title (GTK_WINDOW (window), line);

    gtk_container_border_width (GTK_CONTAINER (window), 4);
    vbox = gtk_vbox_new (TRUE, 0);
    gtk_container_add (GTK_CONTAINER(window), vbox);
    gtk_container_border_width (GTK_CONTAINER (vbox), 2);

#ifdef USE_PIXMAP
    style = gtk_widget_get_style(window);
    pixmap = gdk_pixmap_create_from_xpm_d (window->window, &mask,
                                           &style->bg[GTK_STATE_NORMAL],
                                           (gchar **)PIXMAP_NAME);
#endif
    for(i= 0; i < nr_slot; i++) {
	if (i) {
	    slotFrame[i].sep = gtk_hseparator_new ();
	    gtk_box_pack_start (GTK_BOX (vbox), slotFrame[i].sep, FALSE, TRUE, 0);
	    gtk_widget_show (slotFrame[i].sep);
	}
	sprintf(slotFrame[i].num, "%d", i);
	slotFrame[i].hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), slotFrame[i].hbox, TRUE, FALSE, 0);
#ifdef USE_PIXMAP
	slotFrame[i].pix = gtk_pixmap_new(pixmap, mask);
	gtk_widget_show(slotFrame[i].pix);
	slotFrame[i].pwrbutton = gtk_toggle_button_new();
	gtk_container_add( GTK_CONTAINER(slotFrame[i].pwrbutton), slotFrame[i].pix);
#else
	sprintf(line, " %d ", i);
	slotFrame[i].pwrbutton = gtk_toggle_button_new_with_label(line);
#endif
	gtk_signal_connect (GTK_OBJECT (slotFrame[i].pwrbutton), "toggled",
			    GTK_SIGNAL_FUNC (toggle_buttonCB), (gpointer)slotFrame[i].num);

	gtk_box_pack_start(GTK_BOX(slotFrame[i].hbox), slotFrame[i].pwrbutton, FALSE, FALSE, 0);
	gtk_widget_show (slotFrame[i].pwrbutton);

	slotFrame[i].drvbox = gtk_hbox_new (FALSE, 0);
	gtk_container_border_width (GTK_CONTAINER (slotFrame[i].drvbox), 0);
	sprintf(line, " %s ",cardInfo[i].drv);
	slotFrame[i].drvlabel = gtk_label_new (line);
	gtk_box_pack_start (GTK_BOX (slotFrame[i].drvbox), slotFrame[i].drvlabel,
			    FALSE, FALSE, 3);
	gtk_widget_show (slotFrame[i].drvlabel);
	gtk_widget_show (slotFrame[i].drvbox);
	slotFrame[i].drvbutton = gtk_button_new ();
	gtk_container_add  (GTK_CONTAINER (slotFrame[i].drvbutton), slotFrame[i].drvbox);
	gtk_box_pack_start (GTK_BOX(slotFrame[i].hbox), slotFrame[i].drvbutton,
			    FALSE, FALSE, 1);
	gtk_widget_show (slotFrame[i].drvbutton);

	if (!small)
	    slotFrame[i].label = gtk_label_new (cardInfo[i].name);
	if (cardInfo[i].stat == 2) {
	    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (slotFrame[i].pwrbutton), TRUE);
	    if (sensitive)
		gtk_widget_set_sensitive(slotFrame[i].drvbutton, FALSE);
	}
	if (!small) {
	    gtk_box_pack_start (GTK_BOX(slotFrame[i].hbox), slotFrame[i].label,
				TRUE, TRUE, 0);
	    gtk_widget_show (slotFrame[i].label);
	}
	slotFrame[i].tooltips = gtk_tooltips_new ();

	sprintf(line, "Slot %d Power ON/OFF", i);
 	gtk_tooltips_set_tip (slotFrame[i].tooltips, slotFrame[i].pwrbutton, line, NULL);
	slotFrame[i].tooltips2 = gtk_tooltips_new ();
	sprintf(line, "( Slot%d ) flags 0x%03x port 0x%3x size %d bytes ",
	    i, cardIO[i].flags, cardIO[i].start, cardIO[i].size);
 	gtk_tooltips_set_tip (slotFrame[i].tooltips2, slotFrame[i].drvbutton, line, NULL);
	gtk_widget_show (slotFrame[i].hbox);
    }
    inputtag = gdk_input_add (SOCKET, (GdkInputCondition)GDK_INPUT_READ,
			      (GdkInputFunction)statChanged, NULL);
    gtk_widget_show (vbox);
#ifndef USE_PIXMAP
    gtk_widget_show (window);
#endif
    gtk_main ();
    return 0;
}
